from __future__ import annotations

import json
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any

import ormsgpack
from langchain_core.runnables import RunnableConfig
from langgraph.checkpoint.base import (
    BaseCheckpointSaver,
    ChannelVersions,
    Checkpoint,
    CheckpointMetadata,
    CheckpointTuple,
    get_checkpoint_id,
)
from langgraph.checkpoint.serde.jsonplus import _msgpack_ext_hook_to_json

from ..common.kusto_client import KustoClient
from ..store.kql_builder import serialize_value


@dataclass(slots=True)
class KustoCheckpointConfig:
    client: KustoClient
    table_name: str = "LangGraphCheckpoints"


class KustoCheckpointSaver(BaseCheckpointSaver[str]):
    def __init__(self, *, config: KustoCheckpointConfig) -> None:
        super().__init__()
        self._client = config.client
        self._table_name = config.table_name
        self._raw_table_name = f"{config.table_name}Raw"
        self._writes_table_name = f"{config.table_name}Writes"
        self._writes_raw_table_name = f"{config.table_name}WritesRaw"

    def _insert_checkpoint_row(
        self,
        thread_id: str,
        checkpoint_ns: str,
        checkpoint_id: str,
        parent_checkpoint_id: str,
        snapshot: dict,
        deleted: bool = False,
    ) -> None:
        """Insert a row into the raw checkpoint table using set-or-append."""
        created_at = datetime.now(timezone.utc).isoformat()

        command = f""".set-or-append {self._raw_table_name} <|
print 
    ThreadId={serialize_value(thread_id)}, 
    CheckpointNamespace={serialize_value(checkpoint_ns)}, 
    CheckpointId={serialize_value(checkpoint_id)}, 
    ParentCheckpointId={serialize_value(parent_checkpoint_id)}, 
    Snapshot={serialize_value(snapshot)}, 
    CreatedAt=todatetime({serialize_value(created_at)}), 
    Deleted={serialize_value(deleted)}"""

        self._client.execute_command(command)

    def _insert_checkpoint_writes_row(
        self,
        thread_id: str,
        checkpoint_ns: str,
        checkpoint_id: str,
        task_id: str,
        writes: Sequence,
    ) -> None:
        """Insert a row into the raw checkpoint writes table."""
        created_at = datetime.now(timezone.utc).isoformat()

        command = f""".set-or-append {self._writes_raw_table_name} <|
print 
    ThreadId={serialize_value(thread_id)}, 
    CheckpointNamespace={serialize_value(checkpoint_ns)}, 
    CheckpointId={serialize_value(checkpoint_id)}, 
    TaskId={serialize_value(task_id)}, 
    Writes={serialize_value(writes)}, 
    CreatedAt=todatetime({serialize_value(created_at)})"""

        self._client.execute_command(command)

    def get_tuple(self, config: RunnableConfig) -> CheckpointTuple | None:
        thread_id = config["configurable"].get("thread_id")
        checkpoint_ns = config["configurable"].get("checkpoint_ns", "")
        checkpoint_id = get_checkpoint_id(config)

        if not thread_id:
            return None

        # Build query to get the checkpoint
        if checkpoint_id:
            # Get specific checkpoint
            query = f"""
            {self._table_name}()
            | where ThreadId == '{thread_id}'
                and CheckpointNamespace == '{checkpoint_ns}'
                and CheckpointId == '{checkpoint_id}'
            | take 1
            """
        else:
            # Get latest checkpoint for the thread
            query = f"""
            {self._table_name}()
            | where ThreadId == '{thread_id}'
                and CheckpointNamespace == '{checkpoint_ns}'
            | top 1 by CreatedAt desc
            """

        result = self._client.execute_query(query)
        if not result or not result.primary_results or len(result.primary_results[0]) == 0:
            return None

        row = result.primary_results[0][0]

        # Deserialize the checkpoint using serde
        snapshot_data = row["Snapshot"]
        if isinstance(snapshot_data, str):
            # If it's a string, encode and use serde
            checkpoint = self.serde.loads_typed(("json", snapshot_data.encode("utf-8")))
        elif isinstance(snapshot_data, dict):
            # If it's a dict (from Kusto dynamic), convert to msgpack bytes then use serde
            msgpack_bytes = ormsgpack.packb(snapshot_data)
            checkpoint = self.serde.loads_typed(("msgpack", msgpack_bytes))
        else:
            raise TypeError(f"Unexpected snapshot data type: {type(snapshot_data)}")

        # Query pending writes for this checkpoint
        writes_query = f"""
        {self._writes_table_name}
        | where ThreadId == '{thread_id}'
            and CheckpointNamespace == '{checkpoint_ns}'
            and CheckpointId == '{row["CheckpointId"]}'
        """
        writes_result = self._client.execute_query(writes_query)

        pending_writes = None
        if writes_result and writes_result.primary_results and len(writes_result.primary_results[0]) > 0:
            # Collect all writes from all tasks
            all_writes = []
            for writes_row in writes_result.primary_results[0]:
                writes_data = writes_row["Writes"]
                if writes_data:
                    # Deserialize writes using serde
                    if isinstance(writes_data, str):
                        writes_list = self.serde.loads_typed(("json", writes_data.encode("utf-8")))
                    elif isinstance(writes_data, dict | list):
                        msgpack_bytes = ormsgpack.packb(writes_data)
                        writes_list = self.serde.loads_typed(("msgpack", msgpack_bytes))
                    else:
                        writes_list = writes_data

                    # Convert list of lists back to list of tuples for LangGraph
                    if isinstance(writes_list, list):
                        writes_list = [tuple(item) if isinstance(item, list) else item for item in writes_list]
                    all_writes.extend(writes_list)
            if all_writes:
                pending_writes = all_writes

        # Build metadata
        metadata: CheckpointMetadata = {
            "source": "loop",
            "step": -1,
            "parents": {},
        }
        if row["ParentCheckpointId"]:
            metadata["parents"] = {checkpoint_ns: row["ParentCheckpointId"]}

        # Build configs
        checkpoint_config = {
            "configurable": {
                "thread_id": thread_id,
                "checkpoint_ns": checkpoint_ns,
                "checkpoint_id": row["CheckpointId"],
            }
        }

        parent_config = None
        if row["ParentCheckpointId"]:
            parent_config = {
                "configurable": {
                    "thread_id": thread_id,
                    "checkpoint_ns": checkpoint_ns,
                    "checkpoint_id": row["ParentCheckpointId"],
                }
            }

        return CheckpointTuple(
            config=checkpoint_config,
            checkpoint=checkpoint,
            metadata=metadata,
            parent_config=parent_config,
            pending_writes=pending_writes,
        )

    def list(
        self,
        config: RunnableConfig | None,
        *,
        filter: dict[str, Any] | None = None,
        before: RunnableConfig | None = None,
        limit: int | None = None,
    ) -> Iterator[CheckpointTuple]:
        thread_id = config["configurable"].get("thread_id") if config else None
        checkpoint_ns = config["configurable"].get("checkpoint_ns", "") if config else ""

        # Build query for checkpoints
        query_parts = [f"{self._table_name}()"]

        # Add filters
        where_clauses = []
        if thread_id:
            where_clauses.append(f"ThreadId == '{thread_id}'")
        if checkpoint_ns:
            where_clauses.append(f"CheckpointNamespace == '{checkpoint_ns}'")

        if where_clauses:
            query_parts.append("| where " + " and ".join(where_clauses))

        # Add before filter
        if before:
            before_checkpoint_id = get_checkpoint_id(before)
            if before_checkpoint_id:
                # Get the timestamp of the "before" checkpoint
                before_query = f"""
                {self._table_name}()
                | where CheckpointId == '{before_checkpoint_id}'
                | project CreatedAt
                | take 1
                """
                before_result = self._client.execute_query(before_query)
                if before_result and before_result.primary_results and len(before_result.primary_results[0]) > 0:
                    before_ts = before_result.primary_results[0][0]["CreatedAt"]
                    query_parts.append(f"| where CreatedAt < datetime({before_ts.isoformat()})")

        # Sort by most recent first
        query_parts.append("| order by CreatedAt desc")

        # Add limit
        if limit:
            query_parts.append(f"| take {limit}")

        checkpoints_query = "\n".join(query_parts)

        # Build query with join to include pending writes
        query = f"""
let checkpoints = {checkpoints_query};
checkpoints
| join kind=leftouter (
    {self._writes_table_name}
    | summarize AllWrites = make_list(Writes) by ThreadId, CheckpointNamespace, CheckpointId
) on ThreadId, CheckpointNamespace, CheckpointId
| project-away ThreadId1, CheckpointNamespace1, CheckpointId1
| order by CreatedAt desc
"""

        result = self._client.execute_query(query)

        if not result or not result.primary_results:
            return

        for row in result.primary_results[0]:
            # Deserialize the checkpoint using serde
            snapshot_data = row["Snapshot"]
            if isinstance(snapshot_data, str):
                checkpoint = self.serde.loads_typed(("json", snapshot_data.encode("utf-8")))
            elif isinstance(snapshot_data, dict):
                msgpack_bytes = ormsgpack.packb(snapshot_data)
                checkpoint = self.serde.loads_typed(("msgpack", msgpack_bytes))
            else:
                checkpoint = snapshot_data

            # Deserialize writes if present using serde
            # AllWrites is a list of 'Writes' entries (one per task), each of which is a list of writes
            raw_writes_list = row["AllWrites"] if "AllWrites" in row else None
            pending_writes = None

            if raw_writes_list:
                all_writes = []
                for writes_data in raw_writes_list:
                    if not writes_data:
                        continue

                    # Deserialize this task's writes
                    if isinstance(writes_data, str):
                        task_writes = self.serde.loads_typed(("json", writes_data.encode("utf-8")))
                    elif isinstance(writes_data, dict | list):
                        msgpack_bytes = ormsgpack.packb(writes_data)
                        task_writes = self.serde.loads_typed(("msgpack", msgpack_bytes))
                    else:
                        task_writes = writes_data

                    # Convert list of lists back to list of tuples for LangGraph
                    if isinstance(task_writes, list):
                        task_writes = [tuple(item) if isinstance(item, list) else item for item in task_writes]

                    all_writes.extend(task_writes)

                if all_writes:
                    pending_writes = all_writes

            # Build metadata
            metadata: CheckpointMetadata = {
                "source": "loop",
                "step": -1,
                "parents": {},
            }
            if row["ParentCheckpointId"]:
                metadata["parents"] = {row["CheckpointNamespace"]: row["ParentCheckpointId"]}

            # Build configs
            checkpoint_config = {
                "configurable": {
                    "thread_id": row["ThreadId"],
                    "checkpoint_ns": row["CheckpointNamespace"],
                    "checkpoint_id": row["CheckpointId"],
                }
            }

            parent_config = None
            if row["ParentCheckpointId"]:
                parent_config = {
                    "configurable": {
                        "thread_id": row["ThreadId"],
                        "checkpoint_ns": row["CheckpointNamespace"],
                        "checkpoint_id": row["ParentCheckpointId"],
                    }
                }

            yield CheckpointTuple(
                config=checkpoint_config,
                checkpoint=checkpoint,
                metadata=metadata,
                parent_config=parent_config,
                pending_writes=pending_writes,
            )

    def put(
        self,
        config: RunnableConfig,
        checkpoint: Checkpoint,
        metadata: CheckpointMetadata,
        new_versions: ChannelVersions,
    ) -> RunnableConfig:
        thread_id = config["configurable"]["thread_id"]
        checkpoint_ns = config["configurable"].get("checkpoint_ns", "")
        checkpoint_id = checkpoint["id"]
        parent_checkpoint_id = config["configurable"].get("checkpoint_id", "")

        # Serialize the checkpoint using serde, then convert to dict for Kusto ingestion
        type_, serialized_bytes = self.serde.dumps_typed(checkpoint)
        if type_ == "msgpack":
            # Convert msgpack bytes to JSON-compatible dict via ormsgpack
            # Use the jsonplus ext_hook to convert extensions to plain JSON
            snapshot = ormsgpack.unpackb(
                serialized_bytes, ext_hook=_msgpack_ext_hook_to_json, option=ormsgpack.OPT_NON_STR_KEYS
            )
        elif type_ == "json":
            snapshot = json.loads(serialized_bytes.decode("utf-8"))
        elif type_ == "null":
            snapshot = {}
        else:
            # Fallback: try to decode as JSON
            snapshot = json.loads(serialized_bytes.decode("utf-8"))

        self._insert_checkpoint_row(
            thread_id=thread_id,
            checkpoint_ns=checkpoint_ns,
            checkpoint_id=checkpoint_id,
            parent_checkpoint_id=parent_checkpoint_id,
            snapshot=snapshot,
        )

        return {
            "configurable": {
                "thread_id": thread_id,
                "checkpoint_ns": checkpoint_ns,
                "checkpoint_id": checkpoint_id,
            }
        }

    def put_writes(
        self,
        config: RunnableConfig,
        writes: Sequence[tuple[str, Any]],
        task_id: str,
        task_path: str = "",
    ) -> None:
        thread_id = config["configurable"]["thread_id"]
        checkpoint_ns = config["configurable"].get("checkpoint_ns", "")
        checkpoint_id = config["configurable"].get("checkpoint_id")

        if not checkpoint_id:
            return

        # Serialize the writes using serde, then convert to dict/list for Kusto ingestion
        type_, serialized_bytes = self.serde.dumps_typed(list(writes))
        if type_ == "msgpack":
            # Convert msgpack bytes to JSON-compatible structure via ormsgpack
            serialized_writes = ormsgpack.unpackb(
                serialized_bytes, ext_hook=_msgpack_ext_hook_to_json, option=ormsgpack.OPT_NON_STR_KEYS
            )
        elif type_ == "json":
            serialized_writes = json.loads(serialized_bytes.decode("utf-8"))
        elif type_ == "null":
            serialized_writes = []
        else:
            # Fallback: try to decode as JSON
            serialized_writes = json.loads(serialized_bytes.decode("utf-8"))

        self._insert_checkpoint_writes_row(
            thread_id=thread_id,
            checkpoint_ns=checkpoint_ns,
            checkpoint_id=checkpoint_id,
            task_id=task_id,
            writes=serialized_writes,
        )

    def delete_thread(self, thread_id: str) -> None:
        # Soft delete: insert a deletion marker
        self._insert_checkpoint_row(
            thread_id=thread_id,
            checkpoint_ns="",
            checkpoint_id="",
            parent_checkpoint_id="",
            snapshot={},
            deleted=True,
        )



from __future__ import annotations

from typing import Iterable

from langgraph.store.base import BaseStore, Op, Result

from langgraph_kusto.store.config import KustoStoreConfig
from langgraph_kusto.store.memory_layer import KustoMemoryLayer
from langgraph_kusto.store.translator import LanggraphOpToKustoOpTranslator

from ..setup_environment import initialize_kusto


class KustoStore(BaseStore):
    def __init__(self, *, config: KustoStoreConfig) -> None:
        self._client = config.client
        self._table_name = config.table_name
        self._embeddings_table_name = config.embeddings_table_name
        self._initialized = False

        self._translator = LanggraphOpToKustoOpTranslator()
        self._memory = KustoMemoryLayer(embedding_fn=config.embedding_function)

    def _ensure_initialized(self) -> None:
        if self._initialized:
            return

        initialize_kusto(
            client=self._client,
            store_table=self._table_name,
            embeddings_table=self._embeddings_table_name,
        )

        self._initialized = True

    async def _a_ensure_initialized(self) -> None:
        if self._initialized:
            return

        initialize_kusto(
            client=self._client,
            store_table=self._table_name,
            embeddings_table=self._embeddings_table_name,
        )

        self._initialized = True

    def batch(self, ops: Iterable[Op]) -> list[Result]:
        self._ensure_initialized()
        results: list[Result] = []

        for op in ops:
            command = self._translator.translate_op(
                op,
                table_name=self._table_name,
                embeddings_table_name=self._embeddings_table_name,
            )

            raw_result = self._memory.execute(command, self._client)
            result = self._translator.translate_result(raw_result, op)
            results.append(result)

        return results

    async def abatch(self, ops: Iterable[Op]) -> list[Result]:
        await self._a_ensure_initialized()
        results: list[Result] = []

        for op in ops:
            command = self._translator.translate_op(
                op,
                table_name=self._table_name,
                embeddings_table_name=self._embeddings_table_name,
            )

            raw_result = await self._memory.aexecute(command, self._client)
            result = self._translator.translate_result(raw_result, op)
            results.append(result)

        return results



from __future__ import annotations

from ..common.kusto_client import KustoClient


class KustoOpenAIEmbeddingFn:
    def __init__(self, *, client: KustoClient, model_uri: str) -> None:
        self._client = client
        self._model_uri = model_uri

    def __call__(self, text: str) -> tuple[list[float], str]:
        escaped = text.replace("'", "''")

        kql = f"evaluate ai_embeddings('{self._model_uri}', '{escaped}')"

        result = self._client.execute_query(kql)
        primary = result.primary_results[0]
        rows = list(primary)
        if not rows:
            raise RuntimeError("No embedding returned from ai_embeddings.")

        row = rows[0]
        value = row["embedding"]
        if isinstance(value, list):
            return [float(x) for x in value], self._model_uri

        # Handle dynamic / JSON-encoded embedding
        return [float(x) for x in value], self._model_uri



from __future__ import annotations

import os

from .common.kusto_client import KustoClient


def initialize_kusto(
    *,
    client: KustoClient | None = None,
    store_table: str | None = None,
    checkpoints_table: str | None = None,
    embeddings_table: str | None = None,
) -> None:
    if client is None:
        client = KustoClient.from_env()

    # Only read from env if not provided (for backward compatibility)
    if store_table is None:
        store_table = os.getenv("KUSTO_STORE_TABLE", "LangGraphStore")
    if checkpoints_table is None:
        checkpoints_table = os.getenv("KUSTO_CHECKPOINTS_TABLE", "LangGraphCheckpoints")
    if embeddings_table is None:
        embeddings_table = os.getenv("KUSTO_STORE_EMBEDDINGS_TABLE", f"{store_table}Embeddings")

    store_raw = f"{store_table}Raw"
    embeddings_raw = f"{embeddings_table}Raw"
    checkpoints_raw = f"{checkpoints_table}Raw"
    checkpoint_writes_raw = f"{checkpoints_table}WritesRaw"

    store_command = (
        f".create table {store_raw} "
        "(Namespace: string, Key: string, Value: dynamic, CreatedAt: datetime, UpdatedAt: datetime, Tags: dynamic, Deleted: bool)"
    )
    embeddings_command = (
        f".create table {embeddings_raw} "
        "(Namespace: string, ParentKey: string, ChunkOrdinal: long, ChunkString: string, Embedding: dynamic, EmbeddingUri: string, CreatedAt: datetime, Deleted: bool)"
    )
    checkpoints_command = (
        f".create table {checkpoints_raw} "
        "(ThreadId: string, CheckpointNamespace: string, CheckpointId: string, ParentCheckpointId: string, Snapshot: dynamic, CreatedAt: datetime, Deleted: bool)"
    )
    checkpoint_writes_command = (
        f".create table {checkpoint_writes_raw} "
        "(ThreadId: string, CheckpointNamespace: string, CheckpointId: string, TaskId: string, Writes: dynamic, CreatedAt: datetime)"
    )

    store_view = (
        f".create-or-alter function with (folder='langgraph') {store_table}() "
        f"{{ {store_raw} | summarize arg_max(UpdatedAt, *) by Namespace, Key | where Deleted == false }}"
    )
    embeddings_view = (
        f".create-or-alter function with (folder='langgraph') {embeddings_table}() "
        f"{{ {embeddings_raw} | summarize arg_max(CreatedAt, *) by Namespace, ParentKey, ChunkOrdinal | where Deleted == false }}"
    )
    checkpoints_view = (
        f".create-or-alter function with (folder='langgraph') {checkpoints_table}() "
        f"{{ {checkpoints_raw} | summarize arg_max(CreatedAt, *) by ThreadId, CheckpointNamespace, CheckpointId | where Deleted == false }}"
    )
    checkpoint_writes_view = (
        f".create-or-alter function with (folder='langgraph') {checkpoints_table}Writes() "
        f"{{ {checkpoint_writes_raw} }}"
    )

    for cmd_name, cmd in [
        ("store table", store_command),
        ("embeddings table", embeddings_command),
        ("checkpoints table", checkpoints_command),
        ("checkpoint writes table", checkpoint_writes_command),
        ("store view", store_view),
        ("embeddings view", embeddings_view),
        ("checkpoints view", checkpoints_view),
        ("checkpoint writes view", checkpoint_writes_view),
    ]:
        try:
            client.execute_command(cmd)
            print(f"Created Kusto {cmd_name}.")
        except Exception as e:
            print(f"Kusto {cmd_name} creation skipped or failed: {e}")


if __name__ == "__main__":
    initialize_kusto()



from __future__ import annotations

import datetime
import json
from typing import Any, Literal

SPECIAL_CHARS = {"\\"}


def _kusto_literal(value: Any) -> str:
    if value is None:
        return "null"
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, str):
        # In KQL, multi-line strings must use triple backticks
        # Also, when potentially troublesome characters are present, just escape with a triple backtick
        if "\n" in value or "\r" in value or any(c in value for c in SPECIAL_CHARS):
            # Triple backtick strings don't need escaping
            return f"```{value}```"
        # Single-line strings use single quotes with doubled quote escaping
        escaped = value.replace("'", "''")
        return f"'{escaped}'"
    if isinstance(value, dict):
        inner = ", ".join(f"{_kusto_literal(k)}: {_kusto_literal(v)}" for k, v in value.items())
        return "{" + inner + "}"
    if isinstance(value, (list, tuple)):
        open_, close_ = ("[", "]") if isinstance(value, list) else ("(", ")")
        body = ", ".join(_kusto_literal(v) for v in value)
        if isinstance(value, tuple) and len(value) == 1:
            body += ","
        return f"{open_}{body}{close_}"
    return repr(value)


def serialize_value(value: Any) -> str:
    """Serialize a Python value to Kusto KQL format."""
    serialized = value
    if isinstance(value, str):
        serialized = f'"{escape_kql_string(value)}"'
    elif isinstance(value, bool):
        # Kusto uses lowercase true/false
        serialized = "true" if value else "false"
    elif isinstance(value, (list, dict)):
        # Serialize lists and dicts as JSON strings
        serialized = f"dynamic({_kusto_literal(value)})"
    elif isinstance(value, datetime.datetime):
        # Format datetime in ISO 8601 format
        serialized = f'datetime("{value.isoformat()}")'
    else:
        serialized = _kusto_literal(value)

    return serialized


def escape_kql_string(value: str) -> str:
    """Escape single quotes for KQL string literals."""
    return value.replace("'", "''")


class KqlBuilder:
    """Builds KQL queries using primitive parameters."""

    @staticmethod
    def memory_get_by_key(
        *, table_name: str, namespace: str, namespace_mode: Literal["prefix", "suffix"], key: str
    ) -> str:
        """Build KQL query to get a single item by key."""
        cond = "startswith" if namespace_mode == "prefix" else "endswith"
        return f"""
{table_name}
| where Namespace {cond} '{namespace}' and Key == '{key}'
| project Value
"""

    @staticmethod
    def memory_get_created_at(
        *, table_name: str, namespace: str, namespace_mode: Literal["prefix", "suffix"], key: str
    ) -> str:
        """Build KQL query to get CreatedAt timestamp for existing item."""
        cond = "startswith" if namespace_mode == "prefix" else "endswith"
        return f"""
{table_name}()
| where Namespace {cond} '{namespace}' and Key == '{key}'
| take 1
| project CreatedAt
"""

    @staticmethod
    def memory_embedding_get_created_at(
        *,
        embeddings_table_name: str,
        namespace: str,
        namespace_mode: Literal["prefix", "suffix"],
        parent_key: str,
        ordinal: int,
    ) -> str:
        """Build KQL query to get CreatedAt timestamp for existing embedding chunk."""
        cond = "startswith" if namespace_mode == "prefix" else "endswith"
        return f"""
{embeddings_table_name}()
| where Namespace {cond} '{namespace}' and ParentKey == '{parent_key}' and ChunkOrdinal == {ordinal}
| take 1
| project CreatedAt
"""

    @staticmethod
    def memory_search_by_similarity(
        *,
        table_name: str,
        embeddings_table_name: str,
        namespace: str,
        namespace_mode: Literal["prefix", "suffix"],
        query_vector: list[float],
        limit: int,
    ) -> str:
        """Build KQL query for vector similarity search."""
        query_vector_json = json.dumps(query_vector)
        cond = "startswith" if namespace_mode == "prefix" else "endswith"

        return f"""
let q = dynamic({query_vector_json});
let store =
    {table_name}
    | where Namespace {cond} '{namespace}';
let emb =
    {embeddings_table_name}
    | where Namespace {cond} '{namespace}';
emb
| extend Score = series_cosine_similarity(Embedding, q)
| top {limit} by Score desc
| summarize arg_max(Score, * ) by ParentKey
| join kind=inner (store) on $left.ParentKey == $right.Key
| project Namespace, Key, Value, Tags, ChunkString, ChunkOrdinal, EmbeddingUri, Score
"""

    @staticmethod
    def memory_search_by_content(
        *, table_name: str, namespace: str, namespace_mode: Literal["prefix", "suffix"], query: str, limit: int
    ) -> str:
        """Build KQL query for text-based content search."""
        query_escaped = query.replace("'", "''")
        cond = "startswith" if namespace_mode == "prefix" else "endswith"

        return f"""
{table_name}()
| where Namespace {cond} '{namespace}'
| where tostring(Value) has '{query_escaped}'
| take {limit}
| project Namespace, Key, Value, Tags, CreatedAt, UpdatedAt
"""

    @staticmethod
    def memory_list_namespaces(*, table_name: str) -> str:
        """Build KQL query to list distinct namespaces."""
        return f"""
{table_name}()
| distinct Namespace
"""



from dataclasses import dataclass
from typing import Any, Callable

from langgraph_kusto.common.kusto_client import KustoClient

# returns a tuple of (embedding vector, metadata)
EmbeddingFunction = Callable[[Any], tuple[list[float], str]]


@dataclass(slots=True)
class KustoStoreConfig:
    client: KustoClient
    table_name: str = "LangGraphStore"
    embeddings_table_name: str = "LangGraphStoreEmbeddings"
    embedding_function: EmbeddingFunction | None = None



from __future__ import annotations

import json
from datetime import datetime, timezone
from typing import Any, Literal, cast

from langgraph.store.base import GetOp, Item, ListNamespacesOp, Op, PutOp, Result, SearchItem, SearchOp

from .memory_ops import MemoryGet, MemoryListNamespaces, MemoryOp, MemoryPut, MemorySearch


class LanggraphOpToKustoOpTranslator:
    """Translates LangGraph operations to Kusto commands and results back to LangGraph items."""

    @staticmethod
    def _namespace_to_str(namespace: tuple[str, ...]) -> str:
        """Convert namespace tuple to string representation."""
        return "/".join(namespace)

    @staticmethod
    def _str_to_namespace(namespace: str | None) -> tuple[str, ...]:
        """Convert string namespace to tuple representation."""
        if not namespace:
            return tuple()
        return tuple(str(namespace).split("/"))

    def translate_op(
        self,
        op: Op,
        *,
        table_name: str,
        embeddings_table_name: str,
        namespace_mode: Literal["prefix", "suffix"] = "prefix",
    ) -> MemoryOp:
        """Translate a LangGraph Op to a memory command."""
        if isinstance(op, GetOp):
            return MemoryGet(
                namespace_match_type=namespace_mode,
                namespace=self._namespace_to_str(op.namespace),
                key=op.key,
                table_name=table_name,
            )
        if isinstance(op, PutOp):
            return MemoryPut(
                namespace=self._namespace_to_str(op.namespace),
                key=op.key,
                value=op.value,
                tags=None,
                table_name=table_name,
                embeddings_table_name=embeddings_table_name,
                namespace_match_type=namespace_mode,
                index=op.index,
            )
        if isinstance(op, SearchOp):
            return MemorySearch(
                namespace=self._namespace_to_str(op.namespace_prefix),
                namespace_match_type=namespace_mode,
                query=op.query,
                limit=op.limit,
                offset=op.offset,
                table_name=table_name,
                embeddings_table_name=embeddings_table_name,
            )
        if isinstance(op, ListNamespacesOp):
            return MemoryListNamespaces(
                match_conditions=op.match_conditions,
                max_depth=op.max_depth,
                limit=op.limit,
                offset=op.offset,
                table_name=table_name,
            )
        raise TypeError(f"Unsupported op type: {type(op)}")

    def translate_get_result(self, raw: Any, op: GetOp) -> Item | None:
        """Translate a raw memory get result to a LangGraph Item."""
        if raw is None:
            return None

        # ``raw`` is expected to be a Kusto SDK result; tests treat it as
        # ``result.primary_results[0][0].to_dict()``.
        if isinstance(raw, list) and raw:
            # Assume first row is the payload
            row = raw[0]
            if hasattr(row, "to_dict"):
                row = row.to_dict()
        else:
            row = raw

        if not isinstance(row, dict):
            return None

        value_raw = row.get("Value")
        if isinstance(value_raw, str):
            try:
                value = json.loads(value_raw)
            except Exception:  # pragma: no cover - defensive
                value = {"value": value_raw}
        elif isinstance(value_raw, dict):
            value = cast(dict[str, Any], value_raw)
        else:
            value = {"value": value_raw}

        created_at = row.get("CreatedAt") or datetime.now(timezone.utc)
        updated_at = row.get("UpdatedAt") or created_at

        return Item(
            value=value,
            key=op.key,
            namespace=op.namespace,
            created_at=created_at,
            updated_at=updated_at,
        )

    def translate_put_result(self, raw: Any, op: PutOp) -> None:
        """Translate a raw memory put result (always returns None)."""
        return None

    def translate_search_result(self, raw: list[dict], op: SearchOp) -> list[SearchItem]:
        """Translate raw memory search results to LangGraph SearchItems."""
        items: list[SearchItem] = []
        # ``raw`` is expected to be a list of dicts from ``row.to_dict()``.
        for row in raw:
            ns_tuple = self._str_to_namespace(cast(str | None, row.get("Namespace")))
            value_raw = row.get("Value")
            score = cast(float | None, row.get("Score"))

            if isinstance(value_raw, str):
                try:
                    value = json.loads(value_raw)
                except Exception:  # pragma: no cover - defensive
                    value = {"value": value_raw}
            elif isinstance(value_raw, dict):
                value = cast(dict[str, Any], value_raw)
            else:
                value = {"value": value_raw}

            created_at = row.get("CreatedAt") or datetime.now(timezone.utc)
            updated_at = row.get("UpdatedAt") or created_at

            items.append(
                SearchItem(
                    namespace=ns_tuple,
                    key=cast(str, row.get("Key")),
                    value=value,
                    created_at=created_at,
                    updated_at=updated_at,
                    score=score,
                )
            )

        return items

    def translate_list_namespaces_result(self, raw: list[str], op: ListNamespacesOp) -> list[tuple[str, ...]]:
        """Translate raw memory namespace list to LangGraph namespace tuples."""
        namespaces = [self._str_to_namespace(ns) for ns in raw]

        def matches(ns: tuple[str, ...]) -> bool:
            if op.match_conditions is None:
                return True
            for cond in op.match_conditions:
                path = cond.path
                if cond.match_type == "prefix":
                    if tuple(ns[: len(path)]) != path:
                        return False
                else:
                    if tuple(ns[-len(path) :]) != path:
                        return False
            return True

        filtered: list[tuple[str, ...]] = []
        for ns in namespaces:
            if not matches(ns):
                continue
            if op.max_depth is not None and len(ns) > op.max_depth:
                ns = ns[: op.max_depth]
            filtered.append(ns)

        return filtered

    def translate_result(self, raw: Any, op: Op) -> Result:
        """Translate a raw memory result to the appropriate LangGraph result type."""
        if isinstance(op, GetOp):
            return self.translate_get_result(raw, op)
        if isinstance(op, PutOp):
            return self.translate_put_result(raw, op)
        if isinstance(op, SearchOp):
            return self.translate_search_result(raw, op)
        if isinstance(op, ListNamespacesOp):
            return self.translate_list_namespaces_result(raw, op)
        raise TypeError(f"Unsupported op type: {type(op)}")



from __future__ import annotations

import datetime
import json
import re
from typing import Any, Callable, cast

from ..common import utc_now
from ..common.kusto_client import KustoClient
from .config import EmbeddingFunction
from .kql_builder import KqlBuilder, escape_kql_string, serialize_value
from .memory_ops import MemoryGet, MemoryListNamespaces, MemoryOp, MemoryPut, MemorySearch


class KustoMemoryLayer:
    """Executes memory commands by generating KQL and calling the Kusto client.

    Responsibilities:
    - Execute abstract memory ops against Kusto (Get, Put, Search, List namespaces)
    - Handle embedding generation when embedding_fn is provided
    - Manage dual-table pattern (raw table + embeddings table)
    - Preserve CreatedAt semantics for both raw data and embedding chunks
    """

    def __init__(self, *, embedding_fn: EmbeddingFunction | None = None) -> None:
        """Initialize the Kusto Memory Layer.

        Parameters:
            embedding_fn: Optional function to generate embeddings from content.
                         Takes content and returns (vector, metadata) tuple.
        """
        self._embedding_fn = embedding_fn

    @staticmethod
    def _parse_json_path(path: str) -> list[str | int]:
        """Parse a JSON path string into a list of keys and indices.

        Supports:
        - Simple fields: "field"
        - Nested fields: "parent.child"
        - Array indexing: "array[0]", "array[-1]", "array[*]"

        Examples:
        - "metadata.title" -> ["metadata", "title"]
        - "context[*].content" -> ["context", "*", "content"]
        - "authors[0].name" -> ["authors", 0, "name"]
        """
        keys: list[str | int] = []
        for segment in path.split("."):
            match = re.match(r"(.+)\[(.*)\]$", segment)
            if match:
                name, index = match.groups()
                keys.append(name)
                if index == "*":
                    keys.append("*")
                else:
                    keys.append(int(index))
            else:
                keys.append(segment)
        return keys

    @staticmethod
    def _traverse_json_path(data: Any, keys: list[str | int]) -> list[Any]:
        """Traverse a data structure using parsed JSON path keys.

        Returns a list of values found at the specified path.
        Handles wildcards (*) by expanding to all array elements.
        """
        if not keys:
            return [data]

        key = keys[0]
        remaining = keys[1:]

        if key == "*":
            if not isinstance(data, list):
                return []
            results: list[Any] = []
            for item in data:
                results.extend(KustoMemoryLayer._traverse_json_path(item, remaining))
            return results

        if isinstance(key, int):
            if not isinstance(data, list):
                return []
            try:
                item = data[key]
                return KustoMemoryLayer._traverse_json_path(item, remaining)
            except IndexError:
                return []

        # key is string
        if not isinstance(data, dict):
            return []

        if key not in data:
            return []

        return KustoMemoryLayer._traverse_json_path(data[key], remaining)

    @staticmethod
    def _extract_fields(value: Any, paths: list[str]) -> list[tuple[str, str]]:
        """Extract field values from a data structure using JSON paths.

        Returns a list of (path, serialized_value) tuples.
        Each path may produce multiple values if wildcards are used.
        """
        results: list[tuple[str, str]] = []

        for path in paths:
            keys = KustoMemoryLayer._parse_json_path(path)
            values = KustoMemoryLayer._traverse_json_path(value, keys)

            for extracted_value in values:
                if isinstance(extracted_value, str):
                    serialized = extracted_value
                else:
                    serialized = json.dumps(extracted_value)
                results.append((path, serialized))

        return results

    def _enrich_command_with_embeddings(self, command: MemoryPut | MemorySearch) -> None:
        """Enrich Put and Search commands with embeddings when embedding_fn is available.

        For MemoryPut:
        - Handles indexing configuration (None, False, or list[str])
        - Extracts field values based on index configuration
        - Calls embedding_fn to get vector + metadata for each field
        - Populates embedding_chunks with (ordinal, chunk_string, vector)
        - Sets embedding_model_uri from metadata

        For MemorySearch:
        - Calls embedding_fn on the query string
        - Populates query_vector for similarity search
        """
        if self._embedding_fn is None:
            return

        if isinstance(command, MemoryPut):
            # Handle index configuration
            if command.index is False:
                # Explicitly disabled indexing
                return

            if command.index is None:
                # Default behavior: embed the whole value
                serialized_value = json.dumps(command.value)
                (vector, metadata) = self._embedding_fn(serialized_value)
                command.embedding_chunks = [(0, serialized_value, vector)]
                command.embedding_model_uri = metadata

            elif isinstance(command.index, list):
                # Extract and embed specific fields
                field_values = self._extract_fields(command.value, command.index)

                chunks: list[tuple[int, str, list[float]]] = []
                metadata = None

                for ordinal, (path, serialized_field_value) in enumerate(field_values):
                    (vector, meta) = self._embedding_fn(serialized_field_value)
                    chunks.append((ordinal, serialized_field_value, vector))
                    if metadata is None:
                        metadata = meta

                command.embedding_chunks = chunks
                command.embedding_model_uri = metadata

        elif isinstance(command, MemorySearch) and command.query:
            (vector, _) = self._embedding_fn(command.query)
            command.query_vector = vector

    def _ingest_rows(self, client: KustoClient, table: str, rows: list[dict]) -> None:
        """Ingest rows into Kusto table using .set-or-append command."""
        if not rows:
            return

        # Build the print statements for each row
        print_statements = []
        for row in rows:
            # Build column=value pairs using Kusto serialization
            columns = ", ".join([f"{key}={serialize_value(value)}" for key, value in row.items()])
            print_statements.append(f"print {columns}")

        # Join all print statements with union
        union_query = " | union ".join(print_statements)

        # Execute the .set-or-append command
        command = f".set-or-append {table} <| {union_query}"
        client.execute_command(command)

    async def _ingest_rows_async(self, client: KustoClient, table: str, rows: list[dict]) -> None:
        """Ingest rows into Kusto table using .set-or-append command asynchronously."""
        if not rows:
            return

        # Build the print statements for each row
        print_statements = []
        for row in rows:
            # Build column=value pairs using Kusto serialization
            columns = ", ".join([f"{key}={serialize_value(value)}" for key, value in row.items()])
            print_statements.append(f"print {columns}")

        # Join all print statements with union
        union_query = " | union ".join(print_statements)

        # Execute the .set-or-append command
        command = f".set-or-append {table} <| {union_query}"
        await client.execute_command_async(command)

    def execute(self, command: MemoryOp, client: KustoClient) -> Any:
        """Execute a memory command synchronously."""
        # Enrich with embeddings before execution if needed
        if isinstance(command, (MemoryPut, MemorySearch)):
            self._enrich_command_with_embeddings(command)

        if isinstance(command, MemoryGet):
            return self._execute_get(command, client)
        if isinstance(command, MemoryPut):
            return self._execute_put(command, client)
        if isinstance(command, MemorySearch):
            return self._execute_search(command, client)
        if isinstance(command, MemoryListNamespaces):
            return self._execute_list_namespaces(command, client)
        raise TypeError(f"Unsupported command type: {type(command)}")

    async def aexecute(self, command: MemoryOp, client: KustoClient) -> Any:
        """Execute a memory command asynchronously."""
        # Enrich with embeddings before execution if needed
        if isinstance(command, (MemoryPut, MemorySearch)):
            self._enrich_command_with_embeddings(command)

        if isinstance(command, MemoryGet):
            return await self._aexecute_get(command, client)
        if isinstance(command, MemoryPut):
            return await self._aexecute_put(command, client)
        if isinstance(command, MemorySearch):
            return await self._aexecute_search(command, client)
        if isinstance(command, MemoryListNamespaces):
            return await self._aexecute_list_namespaces(command, client)
        raise TypeError(f"Unsupported command type: {type(command)}")

    # ===================== GET =====================

    def _execute_get(self, cmd: MemoryGet, client: KustoClient) -> Any | None:
        """Execute a get command synchronously."""
        query = KqlBuilder.memory_get_by_key(
            namespace_mode=cmd.namespace_match_type,
            table_name=cmd.table_name,
            namespace=cmd.namespace,
            key=cmd.key,
        )
        result = client.execute_query(query)
        table = result.primary_results[0]
        rows = list(table)
        if not rows:
            return None

        row = rows[0]
        if hasattr(row, "to_dict"):
            data = row.to_dict()
        else:
            data = dict(row)

        return data

    async def _aexecute_get(self, cmd: MemoryGet, client: KustoClient) -> Any | None:
        """Execute a get command asynchronously."""
        query = KqlBuilder.memory_get_by_key(
            namespace_mode=cmd.namespace_match_type,
            table_name=cmd.table_name,
            namespace=cmd.namespace,
            key=cmd.key,
        )
        result = await client.execute_query_async(query)
        table = result.primary_results[0]
        rows = list(table)
        if not rows:
            return None

        row = rows[0]
        if hasattr(row, "to_dict"):
            data = row.to_dict()
        else:
            data = dict(row)

        return data

    # ===================== PUT =====================

    def _execute_put(self, cmd: MemoryPut, client: KustoClient) -> None:
        """Execute a put command synchronously."""
        if cmd.value is None:
            now = utc_now()
            raw_table = f"{cmd.table_name}Raw"
            rows = [
                {
                    "Namespace": cmd.namespace,
                    "Key": cmd.key,
                    "Value": {},
                    "CreatedAt": now,
                    "UpdatedAt": now,
                    "Tags": {},
                    "Deleted": True,
                }
            ]
            self._ingest_rows(client, raw_table, rows)
            return None

        self._put_raw(cmd, client)

    async def _aexecute_put(self, cmd: MemoryPut, client: KustoClient) -> None:
        """Execute a put command asynchronously."""
        if cmd.value is None:
            now = utc_now()
            raw_table = f"{cmd.table_name}Raw"
            rows = [
                {
                    "Namespace": cmd.namespace,
                    "Key": cmd.key,
                    "Value": json.dumps(None),
                    "CreatedAt": now,
                    "UpdatedAt": now,
                    "Tags": json.dumps({}),
                    "Deleted": True,
                }
            ]
            await self._ingest_rows_async(client, raw_table, rows)
            return None

        await self._aput_raw(cmd, client)

    def _put_raw(self, cmd: MemoryPut, client: KustoClient) -> None:
        """Put raw data synchronously with embeddings support."""
        now = utc_now()

        created_at = now
        existing_query = KqlBuilder.memory_get_created_at(
            namespace_mode=cmd.namespace_match_type,
            table_name=cmd.table_name,
            namespace=cmd.namespace,
            key=cmd.key,
        )

        result = client.execute_query(existing_query)
        table = result.primary_results[0]
        rows = list(table)
        if rows:
            row = rows[0]
            if hasattr(row, "to_dict"):
                data = row.to_dict()
            else:
                data = dict(row)
            stored_created_at = data.get("CreatedAt")
            if stored_created_at is not None:
                created_at = stored_created_at

        raw_table = f"{cmd.table_name}Raw"
        rows = [
            {
                "Namespace": cmd.namespace,
                "Key": cmd.key,
                "Value": cmd.value,
                "CreatedAt": created_at,
                "UpdatedAt": now,
                "Tags": cmd.tags or {},
                "Deleted": False,
            }
        ]
        self._ingest_rows(client, raw_table, rows)

        self._put_embeddings(cmd, cmd.value, now, client)

    async def _aput_raw(self, cmd: MemoryPut, client: KustoClient) -> None:
        """Put raw data asynchronously with embeddings support."""
        now = utc_now()
        serialized_value = json.dumps(cmd.value)
        serialized_tags = json.dumps(cmd.tags or {})

        created_at = now
        existing_query = KqlBuilder.memory_get_created_at(
            namespace_mode=cmd.namespace_match_type,
            table_name=cmd.table_name,
            namespace=cmd.namespace,
            key=cmd.key,
        )

        result = await client.execute_query_async(existing_query)
        table = result.primary_results[0]
        rows = list(table)
        if rows:
            row = rows[0]
            if hasattr(row, "to_dict"):
                data = row.to_dict()
            else:
                data = dict(row)
            stored_created_at = data.get("CreatedAt")
            if stored_created_at is not None:
                created_at = stored_created_at

        raw_table = f"{cmd.table_name}Raw"
        rows = [
            {
                "Namespace": cmd.namespace,
                "Key": cmd.key,
                "Value": serialized_value,
                "CreatedAt": created_at,
                "UpdatedAt": now,
                "Tags": serialized_tags,
                "Deleted": False,
            }
        ]
        await self._ingest_rows_async(client, raw_table, rows)

        await self._aput_embeddings(cmd, serialized_value, now, client)

    def _put_embeddings(self, cmd: MemoryPut, serialized_value: str, now: Any, client: KustoClient) -> None:
        """Store embeddings synchronously."""
        chunks = cmd.embedding_chunks or []
        if not chunks:
            return

        emb_raw_table = f"{cmd.embeddings_table_name}Raw"
        embedding_rows: list[dict] = []

        for ordinal, chunk_string, vector in chunks:
            chunk_created_at = now
            existing_emb_query = KqlBuilder.memory_embedding_get_created_at(
                namespace_mode=cmd.namespace_match_type,
                embeddings_table_name=cmd.embeddings_table_name,
                namespace=cmd.namespace,
                parent_key=cmd.key,
                ordinal=ordinal,
            )

            emb_result = client.execute_query(existing_emb_query)
            emb_table = emb_result.primary_results[0]
            emb_rows = list(emb_table)
            if emb_rows:
                emb_row = emb_rows[0]
                if hasattr(emb_row, "to_dict"):
                    emb_data = emb_row.to_dict()
                else:
                    emb_data = dict(emb_row)
                stored_chunk_created_at = emb_data.get("CreatedAt")
                if stored_chunk_created_at is not None:
                    chunk_created_at = stored_chunk_created_at

            embedding_rows.append(
                {
                    "Namespace": cmd.namespace,
                    "ParentKey": cmd.key,
                    "ChunkOrdinal": ordinal,
                    "ChunkString": chunk_string.replace('"', "'"),
                    "Embedding": vector,
                    "EmbeddingUri": cmd.embedding_model_uri or "",
                    "CreatedAt": chunk_created_at,
                    "Deleted": False,
                }
            )

        self._ingest_rows(client, emb_raw_table, embedding_rows)

    async def _aput_embeddings(self, cmd: MemoryPut, serialized_value: str, now: Any, client: KustoClient) -> None:
        """Store embeddings asynchronously."""
        chunks = cmd.embedding_chunks or []
        if not chunks:
            return

        emb_raw_table = f"{cmd.embeddings_table_name}Raw"
        embedding_rows: list[dict] = []

        for ordinal, chunk_string, vector in chunks:
            chunk_created_at = now
            existing_emb_query = KqlBuilder.memory_embedding_get_created_at(
                namespace_mode=cmd.namespace_match_type,
                embeddings_table_name=cmd.embeddings_table_name,
                namespace=cmd.namespace,
                parent_key=cmd.key,
                ordinal=ordinal,
            )

            emb_result = await client.execute_query_async(existing_emb_query)
            emb_table = emb_result.primary_results[0]
            emb_rows = list(emb_table)
            if emb_rows:
                emb_row = emb_rows[0]
                if hasattr(emb_row, "to_dict"):
                    emb_data = emb_row.to_dict()
                else:
                    emb_data = dict(emb_row)
                stored_chunk_created_at = emb_data.get("CreatedAt")
                if stored_chunk_created_at is not None:
                    chunk_created_at = stored_chunk_created_at

            vector = cmd.embedding_chunks
            embedding_rows.append(
                {
                    "Namespace": cmd.namespace,
                    "ParentKey": cmd.key,
                    "ChunkOrdinal": ordinal,
                    "ChunkString": chunk_string,
                    "Embedding": vector,
                    "EmbeddingUri": cmd.embedding_model_uri or "",
                    "CreatedAt": chunk_created_at,
                    "Deleted": False,
                }
            )

        await self._ingest_rows_async(client, emb_raw_table, embedding_rows)

    # ===================== SEARCH =====================

    def _execute_search(self, cmd: MemorySearch, client: KustoClient) -> list[dict]:
        """Execute a search command synchronously."""
        query = cmd.query or ""

        if cmd.query_vector is not None and query:
            return self._search_with_embeddings(cmd, query, client)
        return self._search_with_text(cmd, query, client)

    async def _aexecute_search(self, cmd: MemorySearch, client: KustoClient) -> list[dict]:
        """Execute a search command asynchronously."""
        query = cmd.query or ""

        if cmd.query_vector is not None and query:
            return await self._asearch_with_embeddings(cmd, query, client)
        return await self._asearch_with_text(cmd, query, client)

    def _search_with_embeddings(self, cmd: MemorySearch, query: str, client: KustoClient) -> list[dict]:
        """Search using vector embeddings synchronously."""
        query_vector = cmd.query_vector or []

        kql = KqlBuilder.memory_search_by_similarity(
            namespace_mode=cmd.namespace_match_type,
            table_name=cmd.table_name,
            embeddings_table_name=cmd.embeddings_table_name,
            namespace=cmd.namespace,
            query_vector=query_vector,
            limit=cmd.limit,
        )

        result = client.execute_query(kql)
        table = result.primary_results[0]
        rows = list(table)

        items: list[dict] = []
        for row in rows[cmd.offset : cmd.offset + cmd.limit]:
            if hasattr(row, "to_dict"):
                data = row.to_dict()
            else:
                data = dict(row)

            items.append(data)

        return items

    async def _asearch_with_embeddings(self, cmd: MemorySearch, query: str, client: KustoClient) -> list[dict]:
        """Search using vector embeddings asynchronously."""
        query_vector = cmd.query_vector or []
        kql = KqlBuilder.memory_search_by_similarity(
            namespace_mode=cmd.namespace_match_type,
            table_name=cmd.table_name,
            embeddings_table_name=cmd.embeddings_table_name,
            namespace=cmd.namespace,
            query_vector=query_vector,
            limit=cmd.limit,
        )

        result = await client.execute_query_async(kql)
        table = result.primary_results[0]
        rows = list(table)

        items: list[dict] = []
        for row in rows[cmd.offset : cmd.offset + cmd.limit]:
            if hasattr(row, "to_dict"):
                data = row.to_dict()
            else:
                data = dict(row)

            items.append(data)

        return items

    def _search_with_text(self, cmd: MemorySearch, query: str, client: KustoClient) -> list[dict]:
        """Search using text matching synchronously."""
        query_escaped = escape_kql_string(query)
        kql = KqlBuilder.memory_search_by_content(
            table_name=cmd.table_name,
            namespace=cmd.namespace,
            namespace_mode=cmd.namespace_match_type,
            query=query_escaped,
            limit=cmd.limit,
        )

        result = client.execute_query(kql)
        table = result.primary_results[0]
        rows = list(table)

        items: list[dict] = []
        for row in rows[cmd.offset : cmd.offset + cmd.limit]:
            if hasattr(row, "to_dict"):
                data = row.to_dict()
            else:
                data = dict(row)

            items.append(data)

        return items

    async def _asearch_with_text(self, cmd: MemorySearch, query: str, client: KustoClient) -> list[dict]:
        """Search using text matching asynchronously."""
        query_escaped = escape_kql_string(query)
        kql = KqlBuilder.memory_search_by_content(
            table_name=cmd.table_name,
            namespace=cmd.namespace,
            namespace_mode=cmd.namespace_match_type,
            query=query_escaped,
            limit=cmd.limit,
        )

        result = await client.execute_query_async(kql)
        table = result.primary_results[0]
        rows = list(table)

        items: list[dict] = []
        for row in rows[cmd.offset : cmd.offset + cmd.limit]:
            if hasattr(row, "to_dict"):
                data = row.to_dict()
            else:
                data = dict(row)

            items.append(data)

        return items

    # ===================== LIST NAMESPACES =====================

    def _execute_list_namespaces(self, cmd: MemoryListNamespaces, client: KustoClient) -> list[str]:
        """Execute a list namespaces command synchronously."""
        kql = f"""
{cmd.table_name}()
| distinct Namespace
"""

        result = client.execute_query(kql)
        table = result.primary_results[0]
        rows = list(table)

        namespaces: list[str] = []
        for row in rows:
            if hasattr(row, "to_dict"):
                data = row.to_dict()
            else:
                data = dict(row)
            ns = data.get("Namespace")
            if ns:
                namespaces.append(ns)

        return namespaces[cmd.offset : cmd.offset + cmd.limit]

    async def _aexecute_list_namespaces(self, cmd: MemoryListNamespaces, client: KustoClient) -> list[str]:
        """Execute a list namespaces command asynchronously."""
        kql = f"""
{cmd.table_name}()
| distinct Namespace
"""

        result = await client.execute_query_async(kql)
        table = result.primary_results[0]
        rows = list(table)

        namespaces: list[str] = []
        for row in rows:
            if hasattr(row, "to_dict"):
                data = row.to_dict()
            else:
                data = dict(row)
            ns = data.get("Namespace")
            if ns:
                namespaces.append(ns)

        return namespaces[cmd.offset : cmd.offset + cmd.limit]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Literal


@dataclass(slots=True)
class MemoryOp: ...


@dataclass(slots=True)
class MemoryNamespacedOp(MemoryOp):
    namespace: str
    namespace_match_type: Literal["prefix", "suffix"]  # 'prefix' or 'suffix'


@dataclass(slots=True)
class MemoryGet(MemoryNamespacedOp):
    """Command to retrieve a single item from memory store."""

    key: str
    table_name: str


@dataclass(slots=True)
class MemoryPut(MemoryNamespacedOp):
    """Command to store or update an item in memory store."""

    key: str
    value: Any
    tags: dict[str, Any] | None
    table_name: str
    embeddings_table_name: str
    index: list[str] | None | Literal[False] = None
    embedding_chunks: list[tuple[int, str, list[float]]] | None = None
    embedding_model_uri: str | None = None


@dataclass(slots=True)
class MemorySearch(MemoryNamespacedOp):
    """Command to search items in memory store."""

    query: str | None
    limit: int
    offset: int
    table_name: str
    embeddings_table_name: str
    query_vector: list[float] | None = None


@dataclass(slots=True)
class MemoryListNamespaces(MemoryOp):
    match_conditions: tuple[Any, ...] | None
    max_depth: int | None
    limit: int
    offset: int
    table_name: str



from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any

from typing_extensions import Protocol


@dataclass(slots=True)
class KustoConfig:
    cluster_uri: str
    database: str
    default_properties: dict = field(default_factory=dict)


def utc_now() -> datetime:
    return datetime.now(timezone.utc)



from __future__ import annotations

import uuid
from datetime import timedelta
from typing import Any

from azure.identity import (
    AzureCliCredential,
    ChainedTokenCredential,
    EnvironmentCredential,
    InteractiveBrowserCredential,
    ManagedIdentityCredential,
    SharedTokenCacheCredential,
    TokenCachePersistenceOptions,
)
from azure.kusto.data import ClientRequestProperties
from azure.kusto.data import KustoClient as ADXKustoClient
from azure.kusto.data import KustoConnectionStringBuilder

from langgraph_kusto.common import KustoConfig


class KustoClient:
    _client_cache: dict[str, ADXKustoClient] = {}
    _client: ADXKustoClient

    def __init__(self, *, config: KustoConfig) -> None:
        self._config = config

        cached = self._client_cache.get(config.cluster_uri)
        if cached is not None:
            self._client = cached
            return

        # Enable token caching for local development
        # allow_unencrypted_storage=True allows fallback to plaintext cache if encryption fails
        cache_options = TokenCachePersistenceOptions(allow_unencrypted_storage=True)

        # Use a chained credential:
        # 1. SharedTokenCacheCredential - reads from the local MSAL cache (silent, no prompt)
        # 2. EnvironmentCredential, ManagedIdentityCredential, AzureCliCredential - non-interactive fallbacks
        # 3. InteractiveBrowserCredential - interactive login that writes to the cache for next time
        credential = ChainedTokenCredential(
            SharedTokenCacheCredential(cache_persistence_options=cache_options),
            EnvironmentCredential(),
            ManagedIdentityCredential(),
            AzureCliCredential(),
            InteractiveBrowserCredential(
                cache_persistence_options=cache_options,
            ),
        )

        kcsb = KustoConnectionStringBuilder.with_azure_token_credential(
            config.cluster_uri,
            credential,
        )

        client = ADXKustoClient(kcsb)
        self._client_cache[config.cluster_uri] = client
        self._client = client

    @property
    def database(self) -> str:
        return self._config.database

    @classmethod
    def from_env(cls) -> KustoClient:
        cluster_uri = cls._get_required_env("KUSTO_CLUSTER_URI")
        database = cls._get_required_env("KUSTO_DATABASE")

        config = KustoConfig(
            cluster_uri=cluster_uri,
            database=database,
        )
        return cls(config=config)

    @staticmethod
    def _get_required_env(name: str) -> str:
        import os

        value = os.getenv(name)
        if not value:
            raise RuntimeError(f"Environment variable {name} is required for Kusto configuration")
        return value

    def _default_request_properties(self) -> ClientRequestProperties:
        request_properties = ClientRequestProperties()
        request_properties.set_option("servertimeout", timedelta(minutes=2))
        request_properties.set_option("clientRequestId", f"langgraph-kusto-client;{str(uuid.uuid4())}")
        return request_properties

    def execute_query(self, query: str, *, properties: dict | None = None) -> Any:
        request_properties = self._default_request_properties()
        merged = dict(self._config.default_properties)
        if properties is not None:
            merged.update(properties)

        for key, value in merged.items():
            request_properties.set_option(key, value)

        return self._client.execute(self._config.database, query, request_properties)

    def execute_command(self, command: str, *, properties: dict | None = None) -> Any:
        request_properties = self._default_request_properties()
        merged = dict(self._config.default_properties)
        if properties is not None:
            merged.update(properties)

        for key, value in merged.items():
            request_properties.set_option(key, value)

        return self._client.execute(self._config.database, command, request_properties)

    async def execute_query_async(self, query: str, *, properties: dict | None = None) -> Any:
        raise NotImplementedError("Async Kusto query execution is not implemented yet.")

    async def execute_command_async(self, command: str, *, properties: dict | None = None) -> Any:
        raise NotImplementedError("Async Kusto command execution is not implemented yet.")
